"""nodoctest
all.py -- much of sage is imported into this module, so you don't
          have to import everything individually. 
"""

#*****************************************************************************
#
#   SAGE: System for Algebra and Geometry Experimentation    
#
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************



from sage.libs.pari import *
import sage.libs.ntl as ntl

import sys
if sys.version_info[:2] < (2, 4):
    print >>sys.stderr, "Sorry, SAGE requires Python 2.4 or later"
    sys.exit(1)


#print """
#**********************************************************
#*  SAGE Version 0.3                                      *
#*  System for Arithmetic Geometry Experimentation        *
#*  (c) William Stein, 2005                               *
#*  http://modular.fas.harvard.edu/SAGE                   *
#*  Distributed under the terms of the GPL                *
#**********************************************************
#Help: object? -> Documentation about 'object'.
#"""

import os

import sage.rings.integer

from sage.rings.rings import *

from sage.linalg.matrix import *
from sage.linalg.matrix_space import *
from sage.linalg.vector import *
from sage.linalg.vector_space import *
from sage.linalg.berlekamp_massey import berlekamp_massey

from sage.misc.misc import cputime, verbose, set_verbose, get_verbose, \
     version, add, union, uniq, powerset, exists, forall, \
     random_sublist, mul, prod

from sage.modular.modsym.modsym import *
from sage.modular.modsym.heilbronn import *
from sage.modular.modform.modform import *
from sage.modular.abvar.modabvar import *
from sage.modular.cusps import *
from sage.modular.modsym.p1list import *
from sage.rings.rings import *
import sage.rings.integer_ring

from sage.ellcurve.constructor import EllipticCurve
from sage.ellcurve.ell_rational_field import curves as ec_curves, optimal_curves as ec_optimal_curves

from sage.libs.mwrank import mwrank_EllipticCurve, mwrank_MordellWeil, set_precision as mwrank_set_precision
#from sage.libs.hanke import hanke_Matrix, hanke_diagonal_form

from sage.modular.dirichlet import *
from sage.modular.congroup import *
from sage.modular.dims import *
from sage.functions.transcendental import *
from sage.groups.abelian import *
from sage.tables.all import *

from sage.rings.arith import *

from os import system

# maximize memory resources
try:
    import resource   # unix only...
    resource.setrlimit(resource.RLIMIT_AS, (-1,-1))
except:
    pass

# abreviations
EC = EllipticCurve
IM = IntegerModRing
Integers = IM
MF = ModularForms
MS = ModularSymbols
NF = NumberField
PR = PolynomialRing
VS = VectorSpace
true = True
false = False


def CS(x):
    return x.cuspidal_subspace()

def NS(x):
    return x.new_subspace()

def OS(x):
    return x.old_subspace()

def ES(x):
    return x.eisenstein_subspace()

def NCMS(n,k=2,sign=1):
    return ms(n,weight=k,sign=sign).cuspidal_subspace().new_subspace()

tm = misc.cputime

Q = rings.RationalField()
Z = rings.IntegerRing()
R = rings.RealField()
C = rings.ComplexField()
I = C.gen(0)
i = C.gen(0)

QQ = Q
ZZ = Z
RR = R
CC = C

def quo(x):
    return QQ(x)

def rprint(x):
    for z in x:
        print z

def pr(x):
    for z in x:
        print z

from sage.misc.functional import *

##########################################################################
def benchmark(n=-1):
    """
    Run a well-chosen range of SAGE commands and record the time it
    takes for each to run.

    INPUT:
        n -- int (default: -1) the benchmark number; the default
             of -1 runs all the benchmarks.
    OUTPUT:
        list -- summary of timings for each benchmark.
    """
    import sage.misc.benchmark
    return sage.misc.benchmark.benchmark(n)

##########################################################################
# I tried this save/restore session, but it seems hopeless.
# For example, how to restore the local caches of modular symbols
# spaces and vector spaces to the modules?  It's probably better
# to have an object-by-object database anyways.

# def save_session(dirname):
#     """
#     Save the current session to the named directory.  A file is
#     created for each object currently defined.
#     """
#     if os.path.exists(dirname):
#         os.rename(dirname,dirname+".old")
#     os.mkdir(dirname)
#     for x in show_identifiers():
#         db.save(globals()[x], "%s/%s"%(dirname,x))

# import posix
# def load_session(dirname):
#     """
#     Load all variables defined in a session from the named directory.
#     """
#     assert os.path.exists(dirname), "Directory %s does not exist."%dirname
#     for x in posix.listdir(dirname):
#         globals()[x] = db.load("%s/%s"%(dirname,x))

        
__iglob = globals().keys()[:] + \
        ['os', '__', '_i1', '_i', '_ii', '___', 'show_identifiers', '_iii', '_', 'iglob']

def show_identifiers():
    """
    Returns a list of the names of non-hidden (i.e., not starting with
    an underscore) identifiers defined by the user in this session.
    These are the identifiers saved by the save command.
    """
    return [k for k in globals().keys() if not k in __iglob and k[0] != "_"]



# Grab signal handler back from PARI
def __my_sigint(x, n):
    raise KeyboardInterrupt

import signal
signal.signal(signal.SIGINT, __my_sigint)
signal.signal(signal.SIGABRT, __my_sigint)
