"""
Elliptic curves over finite fields
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import ell_generic
import sage.rings.rings

class EllipticCurve_finite_field(ell_generic.EllipticCurve_generic):
    """
    Elliptic curve over a finite field.
    """
    def __init__(self, x, y=None):
        if isinstance(x, list):
            ainvs = x
            field = ainvs[0].parent()
        else:
            field = x
            ainvs = y
        if not (isinstance(field, sage.rings.rings.Ring) and isinstance(ainvs,list)):
            raise TypeError
        
        ell_generic.EllipticCurve_generic.__init__(self, [field(x) for x in ainvs])

    def points(self):
        """
        All the points on this elliptic curve.
        """
        try:
            return self.__points
        except AttributeError: pass
        
        print "TODO: Stupid point enumeration implementation!"
        points = []
        for x in self.base_field():
            for y in self.base_field():
                try:
                    points.append(self(x,y))
                except ArithmeticError:
                    pass
        self.__points = points
        return self.__points

    def __getitem__(self, n):
        return self.points()[n]

if __name__ ==  '__main__':
    from sage.rings.rings import *
    from constructor import EllipticCurve
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
