"""
NTL ZZ Integer class class
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

include 'interrupt.pxi'
include 'misc.pxi'
include 'decl.pxi'

cdef class ntl_ZZ:
    r"""
    The \class{ZZ} class is used to represent signed, arbitrary length integers.

    Routines are provided for all of the basic arithmetic operations, as
    well as for some more advanced operations such as primality testing.
    Space is automatically managed by the constructors and destructors.

    This module also provides routines for generating small primes, and
    fast routines for performing modular arithmetic on single-precision
    numbers.

    \begin{notice}
    SAGE wrapping of this class is currently not complete.
    \end{notice}
    """
    # See ZZ.pxd for definition of data members

    def __dealloc__(self):
        del_ZZ(self.x)

    def __repr__(self):
        _sig_on
        return string(ZZ_to_str(self.x))


    def __mul__(ntl_ZZ self, other):
        cdef ntl_ZZ y
        if not isinstance(other, ntl_ZZ):
            other = ntl_ZZ(other)
        y = other
        _sig_on
        return make_ZZ(ZZ_mul(self.x, y.x))
            
    def __sub__(ntl_ZZ self, other):
        cdef ntl_ZZ y
        if not isinstance(other, ntl_ZZ):
            other = ntl_ZZ(other)
        y = other
        _sig_on
        return make_ZZ(ZZ_sub(self.x, y.x))

    def __add__(ntl_ZZ self, other):
        cdef ntl_ZZ y
        if not isinstance(other, ntl_ZZ):
            other = ntl_ZZ(other)
        y = other
        _sig_on
        return make_ZZ(ZZ_add(self.x, y.x))
            
    def __pow__(ntl_ZZ self, long e, ignored):
        _sig_on
        return make_ZZ(ZZ_pow(self.x, e))

    cdef set(self, void *y):  # only used internally for initialization; assumes self.x not set yet!
        self.x = <ZZ*> y
        
cdef public make_ZZ(ZZ* x):
    cdef ntl_ZZ y
    _sig_off
    y = ntl_ZZ()
    y.x = x
    return y

def make_new_ZZ(x='0'):
    s = str(x)
    cdef ntl_ZZ n
    n = ntl_ZZ()
    _sig_on
    n.x = str_to_ZZ(s)
    _sig_off
    return n
