#include <iostream>
#include <sstream>
using namespace std;

#include "wrap.h"
#include <NTL/mat_poly_ZZ.h>

//////// ZZ //////////

ZZ* new_ZZ() {
  return new ZZ();
}

void del_ZZ(ZZ* n) {
  delete n;
}

ZZ* str_to_ZZ(const char* s) {
  istringstream *out = new istringstream(s);
  ZZ* y = new ZZ();
  *out >> *y;
  delete out;
  return y;
}

/* Todo: could call into the Python C library to do this
   more efficiently, and return a Python string or int directly. 
   This would also vastly reduce the chances of an accidental 
   memory leak.  
   Drawback: Then this wouldn't be a pure C library interface anymore.  */
char* ZZ_to_str(const ZZ* x) 
{
  ostringstream instore;
  instore << (*x);
  int n = strlen(instore.str().data());
  char* buf = new char[n+1];
  strcpy(buf, instore.str().data());
  return buf;
}


struct ZZ* ZZ_mul(const struct ZZ* x, const struct ZZ* y)
{
  ZZ *z = new ZZ();
  mul(*z, *x, *y);
  return z;
}

struct ZZ* ZZ_add(const struct ZZ* x, const struct ZZ* y)
{
  ZZ *z = new ZZ();
  add(*z, *x, *y);
  return z;
}

struct ZZ* ZZ_sub(const struct ZZ* x, const struct ZZ* y)
{
  ZZ *z = new ZZ();
  sub(*z, *x, *y);
  return z;
}

struct ZZ* ZZ_pow(const struct ZZ* x, long e) 
{
  ZZ *z = new ZZ();
  power(*z, *x, e);
  return z;
}

int ZZ_is_one(struct ZZ* x)
{
  return IsOne(*x);
}

int ZZ_is_zero(struct ZZ* x)
{
  return IsZero(*x);
}

//////// ZZX //////////

struct ZZX* ZZX_init()
{
  return new ZZX();
}

ZZX* str_to_ZZX(const char* s) {
  istringstream *out = new istringstream(s);
  ZZX* x = new ZZX();
  *out >> *x;
  delete out;
  return x;
}

char* ZZX_repr(struct ZZX* x)
{
  ostringstream instore;
  instore << (*x);
  int n = strlen(instore.str().data());
  char* buf = new char[n+1];
  strcpy(buf, instore.str().data());
  return buf;
}

void ZZX_dealloc(struct ZZX* x) {
  delete x;
}

struct ZZX* ZZX_copy(struct ZZX* x) {
  return new ZZX(*x);
}

void ZZX_setitem(struct ZZX* x, long i, const char* a)
{
  ZZ* b = str_to_ZZ(a);
  SetCoeff(*x, i, *b);
  free(b);
}

char* ZZX_getitem(struct ZZX* x, long i)
{
  ZZ c = coeff(*x, i);
  return ZZ_to_str(&c);
}

struct ZZX* ZZX_add(struct ZZX* x, struct ZZX* y)
{
  ZZX *z = new ZZX();
  add(*z, *x, *y);
  return z;
}

struct ZZX* ZZX_sub(struct ZZX* x, struct ZZX* y)
{
  ZZX *z = new ZZX();
  sub(*z, *x, *y);
  return z;
}

struct ZZX* ZZX_mul(struct ZZX* x, struct ZZX* y)
{
  ZZX *z = new ZZX();
  mul(*z, *x, *y);
  return z;
}


struct ZZX* ZZX_div(struct ZZX* x, struct ZZX* y, int* divisible)
{
  struct ZZX* z = new ZZX();
  *divisible = divide(*z, *x, *y);
  return z;
}


struct ZZX* ZZX_mod(struct ZZX* x, struct ZZX* y)
{
  struct ZZX* z = new ZZX();
  rem(*z, *x, *y);
  return z;
}



void ZZX_quo_rem(struct ZZX* x, struct ZZX* other, struct ZZX** r, struct ZZX** q)
{
  struct ZZX *qq = new ZZX(), *rr = new ZZX();
  DivRem(*qq, *rr, *x, *other);
  *r = qq; *q = qq;
}


struct ZZX* ZZX_square(struct ZZX* x)
{
  struct ZZX* s = new ZZX();
  sqr(*s, *x);
  return s;
}


int ZZX_equal(struct ZZX* x, struct ZZX* y)
{
  return (*x) == (*y);
}


int ZZX_is_zero(struct ZZX* x)
{
  return IsZero(*x);
}


int ZZX_is_one(struct ZZX* x)
{
  return IsOne(*x);
}


int ZZX_is_monic(struct ZZX* x)
{
  IsOne(LeadCoeff(*x));
}


struct ZZX* ZZX_neg(struct ZZX* x)
{
  struct ZZX* y = new ZZX();
  *y = -*x;
  return y;
}


struct ZZX* ZZX_left_shift(struct ZZX* x, long n)
{
  struct ZZX* y = new ZZX();
  LeftShift(*y, *x, n);
  return y;
}


struct ZZX* ZZX_right_shift(struct ZZX* x, long n)
{
  struct ZZX* y = new ZZX();
  RightShift(*y, *x, n);
  return y;
}


char* ZZX_content(struct ZZX* x)
{
  struct ZZ c;
  content(c, *x);
  return ZZ_to_str(&c);
}


struct ZZX* ZZX_primitive_part(struct ZZX* x)
{
  struct ZZX* p = new ZZX();
  PrimitivePart(*p, *x);
  return p;
}


void ZZX_pseudo_quo_rem(struct ZZX* x, struct ZZX* y, struct ZZX** r, struct ZZX** q)
{
  *r = new ZZX();
  *q = new ZZX();
  PseudoDivRem(**q, **r, *x, *y);
}


struct ZZX* ZZX_gcd(struct ZZX* x, struct ZZX* y)
{
  struct ZZX* g = new ZZX();
  GCD(*g, *x, *y);
  return g;
}


void ZZX_xgcd(struct ZZX* x, struct ZZX* y, struct ZZ** r, struct ZZX** s, \
	      struct ZZX** t, int proof)
{
  *r = new ZZ();
  *s = new ZZX();
  *t = new ZZX();
  XGCD(**r, **s, **t, *x, *y, proof);
}


long ZZX_degree(struct ZZX* x)
{
  return deg(*x);
}


char* ZZX_leading_coefficient(struct ZZX* x)
{
  return ZZ_to_str(&LeadCoeff(*x));
}


char* ZZX_constant_term(struct ZZX* x)
{
  return ZZ_to_str(&ConstTerm(*x));
}


void ZZX_set_x(struct ZZX* x)
{
  SetX(*x);
}


int ZZX_is_x(struct ZZX* x)
{
  return IsX(*x);
}


struct ZZX* ZZX_derivative(struct ZZX* x)
{
  ZZX* d = new ZZX();
  diff(*d, *x);
  return d;
}


struct ZZX* ZZX_reverse(struct ZZX* x)
{
  ZZX* r = new ZZX();
  reverse(*r, *x);
  return r;
}

struct ZZX* ZZX_reverse_hi(struct ZZX* x, int hi)
{
  ZZX* r = new ZZX();
  reverse(*r, *x, hi);
  return r;
}


struct ZZX* ZZX_truncate(struct ZZX* x, long m)
{
  ZZX* t = new ZZX();
  trunc(*t, *x, m);
  return t;
}


struct ZZX* ZZX_multiply_and_truncate(struct ZZX* x, struct ZZX* y, long m)
{
  ZZX* t = new ZZX();
  MulTrunc(*t, *x, *y, m);
  return t;
}


struct ZZX* ZZX_square_and_truncate(struct ZZX* x, long m)
{
  ZZX* t = new ZZX();
  SqrTrunc(*t, *x, m);
  return t;
}


struct ZZX* ZZX_invert_and_truncate(struct ZZX* x, long m)
{
  ZZX* t = new ZZX();
  InvTrunc(*t, *x, m);
  return t;
}


struct ZZX* ZZX_multiply_mod(struct ZZX* x, struct ZZX* y,  struct ZZX* modulus)
{
  ZZX* p = new ZZX();
  MulMod(*p, *x, *y, *modulus);
  return p;
}


struct ZZ* ZZX_trace_mod(struct ZZX* x, struct ZZX* y)
{
  ZZ* p = new ZZ();
  TraceMod(*p, *x, *y);
  return p;
}


char* ZZX_trace_list(struct ZZX* x)
{
  vec_ZZ v;
  TraceVec(v, *x);
  ostringstream instore;
  instore << v;
  int n = strlen(instore.str().data());
  char* buf = new char[n+1];
  strcpy(buf, instore.str().data());
  return buf;
}


struct ZZ* ZZX_resultant(struct ZZX* x, struct ZZX* y, int proof)
{
  ZZ* res = new ZZ();
  resultant(*res, *x, *y, proof);
  return res;
}


struct ZZ* ZZX_norm_mod(struct ZZX* x, struct ZZX* y, int proof)
{
  ZZ* res = new ZZ();
  NormMod(*res, *x, *y, proof);
  return res;
}


struct ZZ* ZZX_discriminant(struct ZZX* x, int proof)
{
  ZZ* d = new ZZ();
  discriminant(*d, *x, proof);
  return d;
}


struct ZZX* ZZX_charpoly_mod(struct ZZX* x, struct ZZX* y, int proof)
{
  ZZX* f = new ZZX();
  CharPolyMod(*f, *x, *y, proof);
  return f;
}


struct ZZX* ZZX_minpoly_mod(struct ZZX* x, struct ZZX* y)
{
  ZZX* f = new ZZX();
  MinPolyMod(*f, *x, *y);
  return f;
}


void ZZX_clear(struct ZZX* x)
{
  clear(*x);
}


void ZZX_preallocate_space(struct ZZX* x, long n)
{
  x->SetMaxLength(n);
}




//////// mat_ZZ //////////

mat_ZZ* new_mat_ZZ(long nrows, long ncols) {
  mat_ZZ* x = new mat_ZZ;
  x->SetDims(nrows, ncols);
  return x;
}

void del_mat_ZZ(mat_ZZ* x) {
  delete x;
}

char* mat_ZZ_to_str(mat_ZZ* x) 
{
  ostringstream instore;
  instore << (*x);
  int n = strlen(instore.str().data());
  char* buf = new char[n+1];
  strcpy(buf, instore.str().data());
  return buf;
}

struct mat_ZZ* mat_ZZ_mul(const struct mat_ZZ* x, const struct mat_ZZ* y)
{
  mat_ZZ *z = new mat_ZZ();
  mul(*z, *x, *y);
  return z;
}

struct mat_ZZ* mat_ZZ_add(const struct mat_ZZ* x, const struct mat_ZZ* y)
{
  mat_ZZ *z = new mat_ZZ();
  add(*z, *x, *y);
  return z;
}

struct mat_ZZ* mat_ZZ_sub(const struct mat_ZZ* x, const struct mat_ZZ* y)
{
  mat_ZZ *z = new mat_ZZ();
  sub(*z, *x, *y);
  return z;
}

struct mat_ZZ* mat_ZZ_pow(const struct mat_ZZ* x, long e) 
{
  mat_ZZ *z = new mat_ZZ();
  power(*z, *x, e);
  return z;
}

long mat_ZZ_nrows(const struct mat_ZZ* x)
{
  return x->NumRows();
}


long mat_ZZ_ncols(const struct mat_ZZ* x)
{
  return x->NumCols();
}

void mat_ZZ_setitem(struct mat_ZZ* x, int i, int j, const struct ZZ* z)
{
  (*x)(i,j) = *z;
}

struct ZZ* mat_ZZ_getitem(const struct mat_ZZ* x, int i, int j)
{
  return new ZZ((*x)(i,j));
}

struct ZZ* mat_ZZ_determinant(const struct mat_ZZ* x, long deterministic)
{
  ZZ* d = new ZZ();
  determinant(*d, *x, deterministic);
  return d;
}

struct mat_ZZ* mat_ZZ_HNF(const struct mat_ZZ* A, const struct ZZ* D)
{
  struct mat_ZZ* W = new mat_ZZ();
  HNF(*W, *A, *D);
  return W;
}


struct ZZX* mat_ZZ_charpoly(const struct mat_ZZ* A)
{
  ZZX* f = new ZZX();
  CharPoly(*f, *A);
  return f;
}

