#ifdef __cplusplus
#include <NTL/ZZ.h>
#include <NTL/ZZX.h>
#include <NTL/mat_ZZ.h>
#include <NTL/HNF.h>
using namespace NTL;
#endif

#ifdef __cplusplus
#define EXTERN extern "C"
#else 
#define EXTERN
#endif


////////  ZZ //////////

#ifndef __cplusplus
struct ZZ;
#endif

EXTERN struct ZZ* new_ZZ(void);
EXTERN void del_ZZ(struct ZZ* x);
EXTERN struct ZZ* str_to_ZZ(const char* s);
EXTERN char* ZZ_to_str(const struct ZZ* x);
EXTERN struct ZZ* ZZ_add(const struct ZZ* x, const struct ZZ* y);
EXTERN struct ZZ* ZZ_sub(const struct ZZ* x, const struct ZZ* y);
EXTERN struct ZZ* ZZ_mul(const struct ZZ* x, const struct ZZ* y);
EXTERN struct ZZ* ZZ_pow(const struct ZZ* x, long e);
EXTERN int is_zero(struct ZZ*x );
EXTERN int is_one(struct ZZ*x );

//////// ZZX //////////
#ifndef __cplusplus
struct ZZX;
#endif

EXTERN struct ZZX* ZZX_init();
EXTERN struct ZZX* str_to_ZZX(const char* s);
EXTERN char* ZZX_repr(struct ZZX* x);
EXTERN void ZZX_dealloc(struct ZZX* x);
EXTERN struct ZZX* ZZX_copy(struct ZZX* x);
EXTERN void ZZX_setitem(struct ZZX* x, long i, const char* a);
EXTERN char* ZZX_getitem(struct ZZX* x, long i);  
EXTERN struct ZZX* ZZX_add(struct ZZX* x, struct ZZX* y);
EXTERN struct ZZX* ZZX_sub(struct ZZX* x, struct ZZX* y);
EXTERN struct ZZX* ZZX_mul(struct ZZX* x, struct ZZX* y);
EXTERN struct ZZX* ZZX_div(struct ZZX* x, struct ZZX* y, int* divisible);
EXTERN struct ZZX* ZZX_mod(struct ZZX* x, struct ZZX* y);
EXTERN void ZZX_quo_rem(struct ZZX* x, struct ZZX* other, struct ZZX** r, struct ZZX** q);
EXTERN struct ZZX* ZZX_square(struct ZZX* x);
EXTERN int ZZX_equal(struct ZZX* x, struct ZZX* y);
EXTERN int ZZX_is_zero(struct ZZX* x);
EXTERN int ZZX_is_one(struct ZZX* x);
EXTERN int ZZX_is_monic(struct ZZX* x);
EXTERN struct ZZX* ZZX_neg(struct ZZX* x);
EXTERN struct ZZX* ZZX_left_shift(struct ZZX* x, long n);
EXTERN struct ZZX* ZZX_right_shift(struct ZZX* x, long n);
EXTERN char* ZZX_content(struct ZZX* x);
EXTERN struct ZZX* ZZX_primitive_part(struct ZZX* x);
EXTERN void ZZX_pseudo_quo_rem(struct ZZX* x, struct ZZX* y, struct ZZX** r, struct ZZX** q);
EXTERN struct ZZX* ZZX_gcd(struct ZZX* x, struct ZZX* y);
EXTERN void ZZX_xgcd(struct ZZX* x, struct ZZX* y, struct ZZ** r, struct ZZX** s, struct ZZX** t, int proof);
EXTERN long ZZX_degree(struct ZZX* x);
EXTERN char* ZZX_leading_coefficient(struct ZZX* x);
EXTERN char* ZZX_constant_term(struct ZZX* x);
EXTERN void ZZX_set_x(struct ZZX* x);
EXTERN int ZZX_is_x(struct ZZX* x);
EXTERN struct ZZX* ZZX_derivative(struct ZZX* x);
EXTERN struct ZZX* ZZX_reverse(struct ZZX* x);
EXTERN struct ZZX* ZZX_reverse_hi(struct ZZX* x, int hi);
EXTERN struct ZZX* ZZX_truncate(struct ZZX* x, long m);
EXTERN struct ZZX* ZZX_multiply_and_truncate(struct ZZX* x, struct ZZX* y, long m);
EXTERN struct ZZX* ZZX_square_and_truncate(struct ZZX* x, long m);
EXTERN struct ZZX* ZZX_invert_and_truncate(struct ZZX* x, long m);
EXTERN struct ZZX* ZZX_multiply_mod(struct ZZX* x, struct ZZX* y,  struct ZZX* modulus);
EXTERN struct ZZ* ZZX_trace_mod(struct ZZX* x, struct ZZX* y);
EXTERN char* ZZX_trace_list(struct ZZX* x);
EXTERN struct ZZ* ZZX_resultant(struct ZZX* x, struct ZZX* y, int proof);
EXTERN struct ZZ* ZZX_norm_mod(struct ZZX* x, struct ZZX* y, int proof);
EXTERN struct ZZ* ZZX_discriminant(struct ZZX* x, int proof);
EXTERN struct ZZX* ZZX_charpoly_mod(struct ZZX* x, struct ZZX* y, int proof);
EXTERN struct ZZX* ZZX_minpoly_mod(struct ZZX* x, struct ZZX* y);
EXTERN void ZZX_clear(struct ZZX* x);
EXTERN void ZZX_preallocate_space(struct ZZX* x, long n);






//////// mat_ZZ //////////

#ifndef __cplusplus
struct mat_ZZ;
#endif

EXTERN struct mat_ZZ* new_mat_ZZ(long nrows, long ncols);
EXTERN void del_mat_ZZ(struct mat_ZZ* x);
EXTERN char* mat_ZZ_to_str(struct mat_ZZ* x);
EXTERN struct mat_ZZ* mat_ZZ_add(const struct mat_ZZ* x, const struct mat_ZZ* y);
EXTERN struct mat_ZZ* mat_ZZ_sub(const struct mat_ZZ* x, const struct mat_ZZ* y);
EXTERN struct mat_ZZ* mat_ZZ_mul(const struct mat_ZZ* x, const struct mat_ZZ* y);
EXTERN struct mat_ZZ* mat_ZZ_pow(const struct mat_ZZ* x, long e);
EXTERN long mat_ZZ_nrows(const struct mat_ZZ* x);
EXTERN long mat_ZZ_ncols(const struct mat_ZZ* x);
EXTERN void mat_ZZ_setitem(struct mat_ZZ* x, int i, int j, const struct ZZ* z);
EXTERN struct ZZ* mat_ZZ_getitem(const struct mat_ZZ* x, int i, int j);
EXTERN struct ZZ* mat_ZZ_determinant(const struct mat_ZZ* x, long deterministic);
EXTERN struct mat_ZZ* mat_ZZ_HNF(const struct mat_ZZ* A, const struct ZZ* D);
EXTERN struct ZZX* mat_ZZ_charpoly(const struct mat_ZZ* A);

