"""
A single element of an ambient space of modular symbols.
"""

import sage.linalg.vector
import sage.misc.misc as misc
import space

class ModularSymbolElement:
    """
    ModularSymbolElement:
    
    A single element of an ambient space of modular symbol.
    """

    def __init__(self, parent, vector):
        if not isinstance(parent, space.ModularSymbolsSpace):
            raise TypeError, "parent must be a ModularSymbolsSpace."
        if not isinstance(vector, sage.linalg.vector.Vector):
            raise TypeError, "vector must be of type Vector."
        self.__parent = parent
        self.__vector = vector
        self.is_vector = True

    def parent(self):
        return self.__parent
        # We grab the space from the cache (or create it), so there
        # are no nasty circular references.  It is very rare that
        # we would ever need this parent anyways...
        #return ModularSymbols(self.__parent)

    def vector(self):
        return self.__vector

    def manin_symbol(self):
        """
        Returns a list of pairs (c, m), where c is a coefficient and m 
        is a Manin symbol, and the sum of the pairs is a representation 
        of this modular symbol in terms of Manin symbols.
        """
        try:
            return self.__manin_symbol
        except AttributeError:
            ms = []
            A = self.parent()
            v = self.vector()
            manin_symbols = A.ambient_space().manin_symbols_basis()
            ms = [(v[i], manin_symbols[i]) for i in \
                  range(v.degree()) if v[i] != 0]
            self.__manin_symbols = ms
        return self.__manin_symbols
    
    def __repr__(self):
        return str(self.__vector)
        try:
            return self.__rep
        except AttributeError:
            symbols = []
            lincomb = []
            for c, sage in self.manin_symbol():
                symbols.append(sage)
                lincomb.append(c)
            self.__rep = misc.repr_lincomb(symbols, lincomb)
        return self.__rep

    def __add__(self, right):
        if self.__parent != right.__parent:
            raise errors.TypeError("parents must be the same")
        return ModularSymbolElement(self.__parent, self.__vector+right.__vector)
    
    def __sub__(self, right):
        if self.__parent != right.__parent:
            raise errors.TypeError("parents must be the same")
        return ModularSymbolElement(self.__parent, self.__vector - right.__vector)
    
    def __mul__(self, right):
        return ModularSymbolElement(self.__parent, self.__vector*right)
    
    def __rmul__(self, left):
        return ModularSymbolElement(self.__parent, self.__vector*left)

    def __neg__(self):
        return ModularSymbolElement(self.__parent, -self.__vector)
    
    def __pos__(self):
        return self

    def __cmp__(self, other):
        if not isinstance(other, ModularSymbolElement):
            if other is 0:
                return self.__vector == 0
            other = self.__parent(other)
        if self.__parent != other.__parent:
            raise TypeError, "parents must be the same"
        return self.__vector.__cmp__(other.__vector)
