"""
Compute Groebner basis via Singular.


sage: import sage.interfaces.gap as gap
sage: Gap = gap.GAP() # please be patient while GAP loads

i="G:=Group((1,2),(1,2,3));"
o = Gap.cmd(i)
o
# _5 = 'Group([ (1,2), (1,2,3) ])'
i="Size(G);"
o = Gap.cmd(i)
o
# _9 = '6'


# Groebner basis via GAP-singular, SAGE example
#  (requires that the GAP "singular" package is
#   installed and configured)

i="LoadPackage(\"singular\");"; o = Gap.cmd(i)
i="R:= PolynomialRing( Rationals, [\"x[0]\",\"x[1]\",\"x[2]\"] );;"; o = Gap.cmd(i)
i="vars:= IndeterminatesOfPolynomialRing(R);;"; o = Gap.cmd(i)
i="x0:= vars[1];; x1:= vars[2];; x2:= vars[3];;"; o = Gap.cmd(i)
i="r:= [ x0*x1*x2 -x0^2*x2, x0^2*x1*x2-x0*x1^2*x2-x0*x1*x2^2, x0*x1-x0*x2-x1*x2];;"; o = Gap.cmd(i)
i="I:= Ideal( R, r );;"; o = Gap.cmd(i)
i="GroebnerBasis( I );"; o = Gap.cmd(i); o
R=MPolynomialRing(QQ,3,"x"); x=R.gens()
exec("L="+o.replace("^","**")); L

## corresponding GAP code (see above)

LoadPackage("singular");
R:= PolynomialRing( Rationals, ["x[0]","x[1]","x[2]"] );;
vars:= IndeterminatesOfPolynomialRing(R);;
x0:= vars[1];; x1:= vars[2];; x2:= vars[3];;
r:= [ x0*x1*x2 -x0^2*x2, x0^2*x1*x2-x0*x1^2*x2-x0*x1*x2^2, x0*x1-x0*x2-x1*x2];;
I:= Ideal( R, r );;
GroebnerBasis( I );

# Groebner basis via Singular, SAGE example

import sage.interfaces.singular as sing
Sing = sing.SING()
i9="ring R=0,(x0,x1,x2),lp; R;"
i10="ideal I1 = [ x0*x1*x2 -x0^2*x2, x0^2*x1*x2-x0*x1^2*x2-x0*x1*x2^2, x0*x1-x0*x2-x1*x2];"
i11="ideal I2=groebner(I1); I2;"
o9 = Sing.cmd(i9); o10 = Sing.cmd(i10); o11 = Sing.cmd(i11)
print o11
#I2[1]=x1^2*x2^2
#I2[2]=x0*x2^3-x1^2*x2^2+x1*x2^3
#I2[3]=x0*x1-x0*x2-x1*x2
#I2[4]=x0^2*x2-x0*x1*x2
s=o11.split("\n")
m=len(s)
ans=list(s[i][6:] for i in range(m)); ans


# corresponding Singular example (as above):
ring R=0,(x0,x1,x2),lp; R;
ideal I1 = [ x0*x1*x2 -x0^2*x2, x0^2*x1*x2-x0*x1^2*x2-x0*x1*x2^2, x0*x1-x0*x2-x1*x2];
ideal I2=groebner(I1); I2;


####

import sage.rings.GB as gb
R=MPolynomialRing(QQ,3,"x"); x=R.gens()
b = [ x[0]*x[1]*x[2] -x[0]^2*x[2], x[0]^2*x[1]*x[2]-x[0]*x[1]^2*x[2]-x[0]*x[1]*x[2]^2, x[0]*x[1]-x[0]*x[2]-x[1]*x[2]]
gb=gb.GBviaGAPsingular(b)



import sage.rings.GB as gb
R=MPolynomialRing(QQ,3,"x"); x=R.gens()
b = [ x[0]*x[1]*x[2] -x[0]^2*x[2], x[0]^2*x[1]*x[2]-x[0]*x[1]^2*x[2]-x[0]*x[1]*x[2]^2, x[0]*x[1]-x[0]*x[2]-x[1]*x[2]]
ans=gb.GBviaSingular(b)

#############################################
"""

import sage.interfaces.gap as gap

def GBviaGAPsingular(self):
    print "\n please be patient while GAP loads\n"
    Gap = gap.GAP() 
    R=self[0].parent()
    x=R.gens()
    n=len(x)
    i="LoadPackage(\"singular\");"
    o = Gap.cmd(i)
    vars=list("\"x["+str(i)+"]\"" for i in range(n))
    i="R:= PolynomialRing( Rationals, "+ str(list("x["+str(i)+"]" for i in range(n)))  +" );;"
    o = Gap.cmd(i)
    i="vars:= IndeterminatesOfPolynomialRing(R);;"
    o = Gap.cmd(i)
    for i in range(n):
        i="x"+str(i)+":= vars["+str(i+1)+"];;"
        o = Gap.cmd(i)
    i="r:="+str(self)
    o = Gap.cmd(i)
    i="I:= Ideal( R, r );;"
    o = Gap.cmd(i)
    i="GroebnerBasis( I );"
    o = Gap.cmd(i)
    o.replace("\n","")
#   exec("L="+o.replace("^","**"))  ### SyntaxError: unqualified exec
                                    ### is not allowed in function 'GB'
                                    ### it contains a nested function
                                    ### with free variables (GB.py, line 67)
#   return L    
    return o.replace("^","**")

import sage.interfaces.singular as sing

def GBviaSingular(self):
    Sing = sing.SING()
    R=self[0].parent()
    x=R.gens()
    n=len(x)
    #vhar=R.characteristic()
    char=0
    i9="ring R=%s,%s,lp; R;" %(char,x)
    i10="ideal I1 = "+ str(self)+";"
    i11="ideal I2=groebner(I1); I2;"
    o9 = Sing.cmd(i9); o10 = Sing.cmd(i10); o11 = Sing.cmd(i11)
    s=o11.split("\n")
    m=len(s)
    ans=list(s[i][6:] for i in range(m))
    for j in range(len(ans)):
        poly=ans[j]
        for i in range(n):
            poly=poly.replace("x"+str(i),"x["+str(i)+"]")
#       ans[j]=exec(poly) ###### returns SyntaxError
        ans[j]=poly
    return ans
