"""
Multivariate polynomial element.
"""

import operator

import arith

from element import Element
from coerce import bin_op

import integer

import polylist
import polydict

class MPolynomial(Element):
    def __init__(self, parent, x):
        Element.__init__(self, parent)
        self.__element = x

    def __repr__(self):
        return "%s"%self.__element

    def __cmp__(self, right):
        if not isinstance(right, Element) or self.parent() != right.parent():
            right = self.parent()(right)
        return self.__element.__cmp__(right.__element)

    def __add__(self, right):
        if not isinstance(right, Element) or self.parent() != right.parent():
            return bin_op(self, right, operator.add)
        return self.parent()(self.__element + right.__element)

    def __sub__(self, right):
        if not isinstance(right, Element) or self.parent() != right.parent():
            return bin_op(self, right, operator.add)
        return self.parent()(self.__element - right.__element)

    def __mul__(self, right):
        if not isinstance(right, Element) or self.parent() != right.parent():
            return bin_op(self, right, operator.mul)
        return self.parent()(self.__element * right.__element)

    def __div__(self, right):
        if not isinstance(right, Element) or self.parent() != right.parent():
            return bin_op(self, right, operator.div)
        return self.parent()(self.__element / right.__element)

    def __neg__(self):
        return self*(-1)

    def __pow__(self, n):
        if not isinstance(n, (int, integer.Integer)):
            raise TypeError, "The exponent (n=%s) must be an integer."%n
        return self.parent()(self.__element**n)

    def __rpow__(self, n):
        if not isinstance(n, (int, integer.Integer)):
            raise TypeError, "The exponent (n=%s) must be an integer."%n
        return self.parent()(self.__element**n)

    def element(self):
        return self.__element


class MPolynomial_polylist(MPolynomial):
    def __init__(self, parent, x):
        if not isinstance(x, polylist.PolyList):
            x = polylist.PolyList(x, 
                                  parent.variable_names(),
                                  parent._monomial_order_function(),
                                  parent.base_ring()(0))
        MPolynomial.__init__(self, parent, x)

    def __repr__(self):
        return self.element().poly_repr()



class MPolynomial_polydict(MPolynomial):
    def __init__(self, parent, x):
        if not isinstance(x, polydict.PolyDict):
            x = polydict.PolyDict(x, parent.base_ring()(0), remove_zero=True)
        MPolynomial.__init__(self, parent, x)

    def __neg__(self):
        return self*(-1)

    def __repr__(self):
        return self.element().poly_repr(self.parent().variable_names(),
                                        atomic_coefficients=self.parent().base_ring().is_atomic_repr())
