"""
Abstract base class for rings
"""

#*****************************************************************************
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************


class Ring:
    """
    Generic ring class, so all the functions below must be defined
    by all rings, since they derive from this class.
    """
    def __init__(self):
        pass
    
    def __repr__(self):
        return "Generic ring"
    
    def __call__(self, x):
        """
        Coerce x into the ring.
        """
        raise NotImplementedError

    def __contains__(self, x):
        """
        True if x is in this ring.
        """
        raise NotImplementedError

    def is_atomic_repr(self):
        """
        True if the elements have atomic string representations, in the sense
        that they print if they print at s, then -s means the negative of s.
        For example, integers are atomic but polynomials are not.
        """
        return False

    def is_ring(self):
        return True
    
    def is_field(self):
        raise NotImplementedError

    def is_finite(self):
        raise NotImplementedError

    def characteristic(self):
        raise NotImplementedError
    
    def name(self):
        raise NotImplementedError
    
    def __hash__(self):
        return hash(self.name())

    def type(self):
        raise NotImplementedError

