"""
This file gathers together all the tables in SAGE.

    * ConwayPolynomials -- database of selected Conway polynomials.

    * CremonaDatabase() - Cremona's tables of elliptic curves and related data.
                
    * Gamma0Wt2Database() -- table of arithmetic informationa about
                    newforms of weight 2 on Gamma_0(N).

    * JonesDatabase() -- returns the John Jones table of number fields
                with bounded ramification and degree <= 6.
                
    * SteinWatkinsDatabase() - The Stein-Watkins tables of elliptic curves
                and related data.

EXAMPLES:
    sage: ConwayPolynomials()
    Frank Luebeck's database of Conway polynomials

    sage: CremonaDatabase()
    Cremona's database of elliptic curves

    sage: Gamma0Wt2Database()
    Table of arithmetic informationa about newforms of weight 2 on Gamma_0(N)

    sage: JonesDatabase()
    John Jones's table of number fields with bounded ramification and degree <= 6

    sage: SteinWatkinsDatabase()
    The Stein-Watkins database of elliptic curves

"""

#*****************************************************************************
#       SAGE: System for Arithmetic Geometry Experimentation
#
#       Copyright (C) 2005 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************



from sage.tables.conway import ConwayPolynomials

from sage.tables.cremona import CremonaDatabase, \
     cremona_letter_code, parse_cremona_label

from sage.tables.gamma0wt2 import Gamma0Wt2Database

from sage.tables.jones import JonesDatabase

from sage.tables.stein_watkins import SteinWatkinsDatabase, SteinWatkinsData, \
     SteinWatkinsPrimeData
